/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package mindex.processors;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.NavigationProcessor;
import messif.objects.LocalAbstractObject;
import messif.operations.query.DumpObjectsToFilesOperation;
import mindex.MetricIndex;
import mindex.navigation.MIndexPreciseInternalCell;
import mindex.navigation.VoronoiCell;
import mindex.navigation.VoronoiLeafCell;

/**
 * Processor that dumps structure of index on per-level basis.
 * For each level, a file with path-from-root and object pair is created.
 * 
 * NOTICE: Currently the second argument of operation is ignored (it is implemented as false is given, i.e. by leaf nodes).
 * 
 * @author Vlastislav Dohnal, Masaryk University, Brno, Czech Republic, dohnal@fi.muni.cz
 */
public class DumpStructureNavigationProcessor implements NavigationProcessor<DumpObjectsToFilesOperation> {
    
    private final DumpObjectsToFilesOperation operation;
    private final MetricIndex mIndex;
    
    /**
     * Creates new approximate navigation processor for given operation and M-Index. The priority queue is initialized.
     * @param operation approximate kNN operation (should be the same for approximate range)
     * @param mIndex M-Index object with configuration and dynamic cell tree
     * @throws AlgorithmMethodException if the M-Index cannot determine the PP for query object
     */
    public DumpStructureNavigationProcessor(DumpObjectsToFilesOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        this.operation = operation;
        this.mIndex = mIndex;
        
//        // Currently the second argument of operation is ignored
//        if (Boolean.TRUE.equals(operation.getArgument(2)))  // == queryByPivotBalls
//            dumpToFilesByPivots(operation, root, path);
//        else
//            dumpToFilesByLeafNodes(operation, root, path); 

        dumpToFilesByLeafNodes(operation, (MIndexPreciseInternalCell)mIndex.getVoronoiCellTree(), new ArrayList<String>());
        operation.endOperation();
    }

    @Override
    public boolean processStep() throws InterruptedException, AlgorithmMethodException, CloneNotSupportedException {
        return false;
    }

    private void dumpToFilesByLeafNodes(DumpObjectsToFilesOperation operation, MIndexPreciseInternalCell n, List<String> path) {
        path.add("foo");
        int i = 0;
        for (Iterator<Map.Entry<Short, VoronoiCell<LocalAbstractObject>>> iterator = n.getChildNodes(); iterator.hasNext(); i++) {
            final Map.Entry<Short, VoronoiCell<LocalAbstractObject>> entry = iterator.next();
            VoronoiCell child = entry.getValue();
            if (child == null)
                continue;

            //path.set(path.size() - 1, Integer.toString(i+1));     // Ordinal number
            path.set(path.size() - 1, Integer.toString(entry.getKey()));       // Pivot index
            if (child instanceof MIndexPreciseInternalCell) {
                dumpToFilesByLeafNodes(operation, (MIndexPreciseInternalCell)child, path);
            } else {
                MIndexPreciseInternalCell newValidCell = dumpToFilesOnLeafNode(operation, (VoronoiLeafCell) child, path);
                // if the leaf cell is not valid any more (it was split)
                if (newValidCell != null)
                    dumpToFilesByLeafNodes(operation, newValidCell, path);
            }
        }
        path.remove(path.size() - 1);
    }

    private MIndexPreciseInternalCell dumpToFilesOnLeafNode(DumpObjectsToFilesOperation operation, VoronoiLeafCell leaf, List<String> path) {
        leaf.readLock();
        // check whether the leaf node wasn't split in the meanwhile
        if (! leaf.isValid()) {
            leaf.readUnLock();
            return (MIndexPreciseInternalCell) leaf.getSubtitutionOfThisLeaf();
        } else {
            operation.addToAnswer(path.size(), path, leaf.getAllObjects());
            leaf.readUnLock();
            return null;
        }
    }
    
    @Override
    public DumpObjectsToFilesOperation getOperation() {
        return operation;
    }

    @Override
    public void close() {
        if (! operation.isFinished())
            operation.endOperation();
    }

    @Override
    public boolean isFinished() {
        return operation.isFinished();
    }

    @Override
    public int getProcessedCount() {
        return mIndex.getObjectCount();
    }

    @Override
    public int getRemainingCount() {
        return 0;
    }
    
}
