/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package mindex.processors;

import java.util.List;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.NavigationProcessor;
import messif.objects.LocalAbstractObject;
import messif.objects.PrecomputedDistancesFilter;
import messif.objects.impl.ObjectFeature;
import messif.objects.impl.ObjectFeatureQuantized;
import messif.operations.AbstractOperation;
import messif.operations.data.BulkInsertOperation;
import messif.operations.query.QuantizeOperation;
import mindex.MIndexPP;
import mindex.MetricIndex;
import mindex.MetricIndexes;
import mindex.navigation.VoronoiLeafCell;

/**
 * Operation processor for QuantizeOperation that converts an inserted object into its quantized form (ObjectFeatureQuantized)
 * 
 * @author xdohnal
 */
public class QuantizeOperationNavigationProcessor extends InsertOperationNavigationProcessor {

    private final QuantizeOperation operation;
    
    /**
     * Initializes the processor for a single-object insert operation. The object is initialized
     *  by M-Index (set M-Index pivot permutation) and added to the cell map.
     * @param operation single-object insert operation
     * @param mIndex M-Index object
     * @throws AlgorithmMethodException if the initialization goes wrong
     */
    public QuantizeOperationNavigationProcessor(QuantizeOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        super(wrapperBulkInsert(operation), mIndex);
        this.operation = operation;
    }

    private static BulkInsertOperation wrapperBulkInsert(QuantizeOperation oper) {
        return new BulkInsertOperation(oper.getQueryObjects());
    }
    
    /**
     * This method adds the created M-Index key from the filter (MIndexKey) to the {@link ObjectFeature}
     * @param leaf
     * @param objects to be processed
     * @throws messif.algorithms.AlgorithmMethodException
     */
    @Override
    protected void processObjectsOnLeaf(VoronoiLeafCell leaf, List<LocalAbstractObject> objects) throws AlgorithmMethodException {
//                throw new AlgorithmMethodException("TODO: rewrite GetMIndexKey using the MetricIndex method.");
        for (LocalAbstractObject obj : objects) {
            MIndexPP mIndexPP = MetricIndexes.readMIndexPP(obj);
            long[] quantizedKeys = new long[] { mIndexPP.getClusterNumber(leaf.getLevel()) };
            LocalAbstractObject quantObj;
            if (obj instanceof ObjectFeature) {
                ObjectFeature feat = (ObjectFeature)obj;
                quantObj = new ObjectFeatureQuantized(feat.getX(), feat.getY(), feat.getOrientation(), feat.getScale(), quantizedKeys);
            } else {
                quantObj = new ObjectFeatureQuantized(0, 0, 0, 0, quantizedKeys);
            }
            quantObj.setObjectKey(obj.getObjectKey());     // Set the original object key
            operation.addToAnswer(quantObj, 0, null);
            obj.unchainFilter((PrecomputedDistancesFilter) mIndexPP);
        }
    }

    @Override
    public void close() {
    }    
 
    /**
     * Factory method for creating an instance of InsertOperationNavigationProcessor for given Bulk/InsertOperation
     * If allowed - initializing the objects (computing distances to pivots) is done in parallel. Acquires write lock on the leaf nodes.
     * @param operation - either instance of InsertOperation or BulkInsertOperation
     * @param mIndex
     * @return initialized instance of InsertOperationNavigationProcessor
     * @throws AlgorithmMethodException 
     */
    public static DataManipulationNavigationProcessor getInstance(AbstractOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        if (operation instanceof QuantizeOperation) {
            return new QuantizeOperationNavigationProcessor((QuantizeOperation)operation, mIndex);
        } else {
            throw new UnsupportedOperationException("QuantizeOperationNavigationProcessor can be created only for QuantizeOperation");
        }
    }
    
}
